/*
 sematree.c Long lived locking utility
 Copyright  2002 by Padraig Brady <Padraig@antefacto.com>.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h> /* Bah, EPERM == EXIT_FAILURE */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#define _GNU_SOURCE
#include <getopt.h>

const char *name=NULL, *op=NULL, *dir="/var/run/sematree";
unsigned long int wait_time=1000, wait_num=0;
enum opt_t {ACQUIRE, RELEASE, INC, DEC} op_e;

void usage(void)
{
    fprintf(stderr, "Usage: sematree {acquire,release,inc,dec} lock_name [OPTION]...\n"
                    "\n"
                    "A \"long lived\" mutex manipulation utility\n"
                    "\n"
                    "  --wait_time=N\n"
                    "     The number of ms to wait between tries to acquire the lock\n"
                    "     The default is 1000 (1 second)\n"
                    "  --wait_number=N\n"
                    "     The number of tries made to acquire the lock\n"
                    "     The default is 0, which means wait forever\n"
                    "  --workdir=DIR\n"
                    "     The (writeable) directory used to maintain the locks\n"
                    "     The default is %s\n"
                    "  --help\n"
                    "     Display this help text\n"
                    "  --version\n"
                    "     Display version information and exit\n",dir);
    exit(EXIT_FAILURE);
}

static struct option const longopts[] =
{
  {"wait_time", required_argument, NULL, 'w'},
  {"wait_number", required_argument, NULL, 'n'},
  {"workdir", required_argument, NULL, 'd'},
  {"version", no_argument, NULL, 'v'},
  {"help", no_argument, NULL, 'h'},
  {NULL, 0, NULL, 0}
};

int file_exists(const char* path) {
    int exists=1;
    int fd = open(path, O_RDONLY);
    if (fd == -1) {
        if (errno == ENOENT)
            exists=0;
    } else {
        close(fd);
    }
    return exists;
}

/* make sure a directory exists */
int create_dir(const char *dir)
{
    struct stat st;
    if (stat(dir, &st) == 0) {
        if (S_ISDIR(st.st_mode)) {
            return 0;
        }
        return ENOTDIR;
    }
    /*
     * EEXIST below catches race here
     */
    if (mkdir(dir, 0777) != 0 && errno != EEXIST) {
        return errno;
    }
    return 0;
}
            
int sematree(void)
{
    unsigned long int num_to_check=wait_num;
    char lock_path_tmp[1024];
    char sema_path_tmp[1024];
    int rc=EXIT_SUCCESS;

    if ((rc=create_dir(dir))!=0) {
        fprintf(stderr,"sematree: error: workdir [%s] problem\n",dir);
        return rc;
    }

    snprintf(lock_path_tmp,sizeof(lock_path_tmp),"%s/%s",dir,name);
    snprintf(sema_path_tmp,sizeof(sema_path_tmp),"%s/%s.sem",dir,name);

    switch (op_e) {
    case ACQUIRE:
        rc=EXIT_FAILURE;
        while(1) {
            int lock_fd = open(lock_path_tmp, O_RDWR|O_CREAT|O_EXCL, S_IRUSR|S_IWUSR|S_IROTH|S_IWOTH);
            if (lock_fd == -1) {
                if (errno == EEXIST) {
                    usleep(wait_time*1000);
                } else {
                    rc = errno;
                    break;
                }
            } else {
                close(lock_fd); /* lock is now held */
                rc = EXIT_SUCCESS;
                break;
            }
            if (wait_num) {
                num_to_check--;
                if (num_to_check==0) {
                    rc = EXIT_FAILURE;
                    break;
                }
            }
        }
        break;
    case RELEASE:
        if (unlink(lock_path_tmp)==-1) {
            if (errno == ENOENT) {
                fprintf(stderr,"sematree: error: lock [%s] was already released!\n",name);
            }
            rc = errno;
        }
        break;
    case INC:
    case DEC:
        if (!file_exists(lock_path_tmp)) {
            fprintf(stderr,"sematree: error: can't inc without acquiring first\n");
            rc = EXIT_FAILURE;
        } else {
            int sem_fd = open(sema_path_tmp, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IROTH|S_IWOTH);
            if (sem_fd == -1) {
                rc = errno;
            } else {
                /* allow space for -9223372036854775807 (20) */
                char strval[21];
                long val;
                memset(strval,0,sizeof(strval));
                if (read(sem_fd,strval,sizeof(strval)-1) == -1) {
                    rc = errno;
                } else {
                    val = strtol(strval, (char**)NULL, 10);
                    op_e == INC ? val++ : val--;
                    snprintf(strval,sizeof(strval),"%ld",val);
                    lseek(sem_fd,0,SEEK_SET);
                    if (write(sem_fd,strval,strlen(strval)) == -1) {
                        rc = errno;
                    }
                }
                if (close(sem_fd)==0) {
                    if (rc == EXIT_SUCCESS)
                        printf("%s\n",strval);
                } else {
                    rc = errno;
                }
                if (val == 0) { /* auto cleanup */
                    unlink(sema_path_tmp); /* remove sema file if present */
                }
            }
        }
        break;
    }
    return rc;
}

int main(int argc, char** argv) {
    
    int optc;

    if (argc < 3) {
        usage();
    }

    op = argv[1];
    name = argv[2];

    if (!strcmp(op,"acquire")) {
        op_e = ACQUIRE;
    } else if (!strcmp(op,"release")) {
        op_e = RELEASE;
    } else if (!strcmp(op,"inc")) {
        op_e = INC;
    } else if (!strcmp(op,"dec")) {
        op_e = DEC;
    } else {
        fprintf(stderr,"Invalid operation [%s]\n", op);
        exit(EXIT_FAILURE);
    }

    while ((optc = getopt_long_only(argc-2, &argv[2], "", longopts, NULL))>0) {
        switch(optc) {
        case 'w':
            wait_time = strtoul(optarg, (char**)NULL, 10);
            break;
        case 'n':
            wait_num = strtoul(optarg, (char**)NULL, 10);
            break;
        case 'd':
            dir = strdup(optarg);
            break;
        case 'v':
            printf("1.00\n");
            exit(EXIT_SUCCESS);
            break;
        case 'h':
        default:
            usage();
            break;
        }
    }

    (void) umask(0); /* any user can get at locks */

    exit(sematree());
}
